# import ptvsd

# ptvsd.enable_attach(address = ('localhost', 5678), redirect_output=True)
# ptvsd.wait_for_attach()
# ptvsd.break_into_debugger()

import os

import numpy as np

from paraview.util.vtkAlgorithm import *
from vtkmodules.numpy_interface import dataset_adapter as dsa
from vtkmodules.vtkCommonCore import vtkDataArraySelection
from vtkmodules.vtkCommonDataModel import (vtkCompositeDataSet,
                                           vtkMultiBlockDataSet,
                                           vtkTable)


def createModifiedCallback(anobject):
    import weakref
    weakref_obj = weakref.ref(anobject)
    anobject = None

    def _markmodified(*args, **kwars):
        o = weakref_obj()
        if o is not None:
            o.Modified()

    return _markmodified


@smproxy.reader(name="PythonEcgReader", label="Python-based Ecg Reader",
                file_description="Ecg", extensions="ecg")
class PythonEcgReader(VTKPythonAlgorithmBase):
    """A reader that reads a Ecg."""

    _standard_egs = ['I', 'II', 'III', 'AVF', 'AVL', 'AVR'] + ['V' + str(n + 1) for n in range(6)]
    _standard_arrays = ['Lead Row', 'Lead Column'] + _standard_egs

    def __init__(self):
        VTKPythonAlgorithmBase.__init__(self, nInputPorts=0, nOutputPorts=1, outputType='vtkMultiBlockDataSet')
        self._filename = None
        self.is_read = False
        self._arrayselection = vtkDataArraySelection()
        self._old_arrayselection = vtkDataArraySelection()
        self._arrayselection.AddObserver("ModifiedEvent", createModifiedCallback(self))
        self._data_read = set()
        self.no_time = True
        self._egs_u = {}
        self._dirname = None
        self.data_time = 0
        self._leads_names = []
        self._ecgs_names = []

    @smproperty.stringvector(name="FileName")
    @smdomain.filelist()
    @smhint.filechooser(extensions="ecg", file_description="Ecg")
    def SetFileName(self, name):
        """Specify filename for the directory to read."""
        if self._filename != name:
            self._filename = name
            self._dirname = os.path.dirname(self._filename)
            self.is_read = False
            self._data_read = set()
            self.no_time = True
            self._egs_u = {}
            self.data_time = 0
            self._leads_names = []
            self._ecgs_names = []
            self._old_arrayselection.DisableAllArrays()
            self.Modified()

    @smproperty.dataarrayselection(name="Leads")
    def GetDataArraySelection(self):
        return self._arrayselection

    def _get_array_selection(self):
        self._get_leads()
        self._arrays_names = self._standard_arrays + self._leads_names
        for name in self._arrays_names:
            if not self._arrayselection.ArrayExists(name):
                self._arrayselection.DisableArray(name)
        return self._arrayselection

    def _read_data(self):
        self._get_ecgs()
        self.is_read = True

    def _get_ecgs(self):
        if any(self._arrayselection.ArrayIsEnabled(field) and field not in self._data_read for field in
               self._ecgs_names):
            egs_name = None
            egs_u = {name: [] for name in self._ecgs_names}
            read_ecgs = {name for name in self._ecgs_names if self._arrayselection.ArrayIsEnabled(name)}

            have_data = False

            with open(os.path.join(self._dirname, 'data_1.ecg'), 'r') as textFile:
                for line in textFile:
                    items = line.split()

                    if items[0] == '*ECG':
                        if not read_ecgs:
                            break
                        have_data = True
                        egs_name = items[1]
                        read_ecgs.discard(egs_name)

                    elif have_data and self._arrayselection.ArrayIsEnabled(
                            egs_name) and egs_name not in self._data_read:
                        egs_u[egs_name].append(float(items[0]))

            # convert to millivolts
            for name in self._ecgs_names:
                if self._arrayselection.ArrayIsEnabled(name) and name not in self._data_read:
                    self._egs_u[name] = np.array(egs_u[name])
                    self._data_read.add(name)


    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def GetTimestepValues(self):
        return self._timesteps

    def _get_update_time(self, outInfo):
        executive = self.GetExecutive()
        if outInfo.Has(executive.UPDATE_TIME_STEP()):
            utime = outInfo.Get(executive.UPDATE_TIME_STEP())
            dtime = self._timesteps[0]
            for atime in self._timesteps:
                if atime > utime:
                    return dtime
                else:
                    dtime = atime
            return dtime
        return self._timesteps[0]

    def _get_leads(self):
        if len(self._leads_names):
            return
        
        leads_names = []
        leads_lbls = []
        leads_xyz = []

        if self._dirname:
            with open(os.path.join(self._dirname, 'bind.bnd'), "r") as textFile:
                for line in textFile:
                    items = line.split()
                    leads_names.append(items[0])
                    leads_lbls.append(items[1:4])
                    leads_xyz.append(items[4:7])

        self._leads_names = leads_names
        self._leads_lbls = np.array(leads_lbls, dtype=int)
        self._leads_xyz = np.array(leads_xyz, dtype=float)
        self._ecgs_names = self._standard_egs + self._leads_names

    def _get_timesteps(self):
        self._nFrames = 0
        if any(self._arrayselection.ArrayIsEnabled(field) for field in self._ecgs_names):
            with open(os.path.join(self._dirname, 'data_1.ecg'), 'r') as textFile:
                for line in textFile:
                    items = line.split()

                    if items[0] == 'SamplesCount':
                        nFrames = int(items[1])
                        if self._nFrames > nFrames or self._nFrames == 0:
                            self._nFrames = nFrames


        self._timesteps = np.arange(max(self._nFrames, 1))

    def RequestInformation(self, request, inInfoVec, outInfoVec):
        if not self.is_read:
            self._get_array_selection()
        self._get_timesteps()
        executive = self.GetExecutive()
        outInfo = outInfoVec.GetInformationObject(0)
        if not self.no_time:
            nSteps = outInfo.Get(executive.TIME_RANGE(), 1)
        if self.no_time or self._nFrames == 0 and nSteps > 2 or self._nFrames > 0 and self._nFrames != nSteps:
            outInfo.Remove(executive.TIME_STEPS())
            outInfo.Remove(executive.TIME_RANGE())

            for t in self._timesteps:
                outInfo.Append(executive.TIME_STEPS(), t)
            outInfo.Append(executive.TIME_RANGE(), self._timesteps[0])
            outInfo.Append(executive.TIME_RANGE(), self._timesteps[-1])
            self.no_time = False
        return 1


    def RequestData(self, request, inInfoVec, outInfoVec):
        self._read_data()
        output1 = vtkMultiBlockDataSet()
        output1_ix = 0

        self.new_data_time = self._get_update_time(outInfoVec.GetInformationObject(0))

        for name in self._ecgs_names:
            self._ecgs_obj = dsa.WrapDataObject(vtkTable())

        for name in self._egs_u:
            self.set_data_enabled(self._ecgs_obj.RowData, name, self._egs_u[name][:len(self._timesteps)])
        if any(self._arrayselection.ArrayIsEnabled(name) and len(self._egs_u[name]) for name in self._ecgs_names):
            output1.SetBlock(output1_ix, self._ecgs_obj.VTKObject)
            output1.GetMetaData(output1_ix).Set(vtkCompositeDataSet.NAME(), 'Cath ECGs')
            output1_ix += 1

        out = vtkMultiBlockDataSet.GetData(outInfoVec, 0)
        if output1_ix:
            out.SetBlock(0, output1)
            out.GetMetaData(0).Set(vtkCompositeDataSet.NAME(), 'ECGs')

        out.GetInformation().Set(out.DATA_TIME_STEP(), self.new_data_time)

        self._old_arrayselection.CopySelections(self._arrayselection)
        self.data_time = self.new_data_time

        return 1

    def set_data_enabled(self, dataset_attrs, arr_name, array=None, attr_name=None, time_dependent=False):
        if attr_name == None:
            attr_name = arr_name
        if self._old_arrayselection.GetArraySetting(arr_name) != self._arrayselection.GetArraySetting(arr_name):
            if self._arrayselection.ArrayIsEnabled(arr_name) and len(array):
                dataset_attrs.append(array, attr_name)
            else:
                dataset_attrs.VTKObject.RemoveArray(attr_name)
        elif time_dependent and self.new_data_time != self.data_time and self._arrayselection.ArrayIsEnabled(
                arr_name) and len(array):
            dataset_attrs.append(array, attr_name)
